@@ -312,6 +312,32 @@ class TCPClientFactory(ClientFactory):
     def clientConnecitonFailed(self, connector, reason):
         pass
     
+class UDPMulticastClient(DatagramProtocol):
+    def __init__(self, client):
+        self.client = client
+        self.port = reactor.listenUDP(0, self)
+        
+    def datagramReceived(self, datagram, address):
+        '''This method is called when a datagram is received. It will interpret
+        the message, and instruct the networkClient object'''
+        
+        # A server tells us that it exists
+        if datagram.startswith('Trosnoth:Server:'):
+            tcpPort = struct.unpack('!I', datagram[16:20])[0]
+            self.client.addServer(datagram[20:], tcpPort, address)
+
+        # We don't recognise this command
+        else:
+            print "Multicast client doesn't understand: ", datagram
+            
+    def askForServers(self):
+        # Ask the network for trosnoth servers.
+        try:
+            self.port.write('Trosnoth:ServerList?', (multicastGroup, multicastPort))
+        except:
+            # If we can't access the network, we can't ask for servers.
+            pass
+
 class UDPClient(DatagramProtocol):
     def __init__(self, client, port=0):
         self.client = client
@@ -821,8 +847,8 @@ class NetworkClient(object):
     '''Talks to the network as a client.'''
 
     def __init__(self, mainInterface):
-        self.udpClient = UDPClient(self, UDPClientPort)
-        self.udpTransmitter = None
+        self.multicastClient = UDPMulticastClient(self)
+        self.udpClient = None
         self.tcpFactory = None
         self.serverDetails = {}
         self.mainInterface = mainInterface
@@ -840,7 +866,7 @@ class NetworkClient(object):
         delay is the number of seconds to wait before assuming that we've
         heard from all servers on the local network.'''
         # First, we broadcast to the local network asking for servers.
-        self.udpClient.askForServers()
+        self.multicastClient.askForServers()
         self.serverDetails = {}
 
         serverListDeferred = defer.Deferred()
@@ -872,10 +898,13 @@ class NetworkClient(object):
         Universe object containing all world details received by the server.
         Returns a deferred, whose callback is passed the Universe object.'''
         result = defer.Deferred()
-        if self.tcpFactory:
+        if self.tcpFactory is not None or self.udpClient is not None:
             result.errback(Exception('Already connected'))
             return result
 
+        # Create a client for this connection.
+        self.udpClient = UDPClient(self, UDPClientPort)
+
         # Look up the server and begin connection.
         ipAddr, tcpPort = self.serverDetails[server]
         try:
@@ -895,10 +924,15 @@ class NetworkClient(object):
         and port.'''
         # Resolve the ip address.
         print 'Resolving host address...'
+        result = defer.Deferred()
+
+        if self.tcpFactory is not None or self.udpClient is not None:
+            result.errback(Exception('Already connected'))
+            return result
+
         reactor.resolve(hostName).addCallbacks(self.gotHostIPAddress,
                                                self.errorGettingHostIP,
                                                [port])
-        result = defer.Deferred()
         print 'Ok.'
         self.connectingDeferred = result
 
@@ -906,6 +940,14 @@ class NetworkClient(object):
 
     def gotHostIPAddress(self, ipAddr, port):
         print 'Host name resolved to %s' % ipAddr
+
+        if self.tcpFactory is not None or self.udpClient is not None:
+            self.connectingDeferred.errback(Exception('Already connected'))
+            return
+
+        # Create a client for this connection.
+        self.udpClient = UDPClient(self, UDPClientPort)
+
         try:
             self.tcpFactory = TCPClientFactory(self, self.udpClient.port)
             reactor.connectTCP(ipAddr, port, self.tcpFactory)
@@ -952,7 +994,6 @@ class NetworkClient(object):
 
         try:
             ipAddr = self.tcpFactory.protocol.transport.addr[0]
-            self.udpTransmitter = UDPClient(self)
 
             # Create a new universe based on the settings.
             halfMapWidth = self.serverSettings['halfMapWidth']
@@ -981,7 +1022,7 @@ class NetworkClient(object):
                             layout.applyTo(curBlock)
 
             # Listen for udp messages.
-            self.udpTransmitter.connect(self.serverSettings['udpPort'], ipAddr)
+            self.udpClient.connect(self.serverSettings['udpPort'], ipAddr)
         except Exception, e:
             logging.writeException()
             self.connectingDeferred.errback(e)
@@ -1020,7 +1061,7 @@ class NetworkClient(object):
         self.tcpFactory = None
         if self.udpClient is not None:
             self.udpClient.port.stopListening()
-        self.udpClient = UDPClient(self, UDPClientPort)
+        self.udpClient = None
 
         # Tell the main interface that connection was lost.
         self.mainInterface.connectionLost()
@@ -1061,7 +1102,7 @@ class NetworkClient(object):
         return defer.succeed(None)
     
     def sendUDP(self, message):
-        self.udpTransmitter.sendDatagram(message)
+        self.udpClient.sendDatagram(message)
 
     def setDetailsInterface(self, interface):
         self.interface = interface
@@ -1182,7 +1223,7 @@ class NetworkClient(object):
         pos1 = struct.pack('f', player.pos[0])
         pos2 = struct.pack('f', player.pos[1])
         message = 'Kill' + player.id + killer.id + shot.id + pos1 + pos2
-        self.udpTransmitter.sendDatagram(message)
+        self.udpClient.sendDatagram(message)
 
     def receivePlayerKilled(self, pId, kId, sId):
         '''Tells the universe that a certain player has been killed, who they
@@ -1206,7 +1247,7 @@ class NetworkClient(object):
         '''Tells the network that a shot has been destroyed (likely by
         colliding with a turretted player)'''
         message = 'KShot' + shot.originatingPlayer.id + shot.id
-        self.udpTransmitter.sendDatagram(message)
+        self.udpClient.sendDatagram(message)
     
     def sendPlayerRespawn(self, ghost):
         '''Tells the network that a player has respawned at the given
@@ -1216,7 +1257,7 @@ class NetworkClient(object):
         pId = ghost.id
         zId = struct.pack('!I', ghost.currentZone.id)
         message = 'Resp' + pId + zId
-        self.udpTransmitter.sendDatagram(message)
+        self.udpClient.sendDatagram(message)
         
     def receivePlayerRespawn(self, pId, zId):
         player = self.world.playerWithId[pId]
@@ -1233,7 +1274,7 @@ class NetworkClient(object):
         numStars = struct.pack('B', stars)
         upgradeType = upgrade.upgradeType
         message = 'Trans' + tId + pId + numStars + upgradeType
-        self.udpTransmitter.sendDatagram(message)
+        self.udpClient.sendDatagram(message)
         
 
     def addStars(self, team, player, stars):
@@ -1242,13 +1283,13 @@ class NetworkClient(object):
         numStars = struct.pack('B', stars)
         pId = player.id
         message = 'AddStars' + tId + pId + numStars
-        self.udpTransmitter.sendDatagram(message)
+        self.udpClient.sendDatagram(message)
 
     def abandonTransaction(self, team):
         '''Informs the network that a transaction has been abandoned by the
         initiating player.'''
         tId = team.id
-        self.udpTransmitter.sendDatagram('Abandon' + tId)
+        self.udpClient.sendDatagram('Abandon' + tId)
 
     def useUpgrade(self, player):
         '''Informs the network that an update has expired, or that
@@ -1257,20 +1298,20 @@ class NetworkClient(object):
         zId = struct.pack('!I', player.currentZone.id)
         message = 'UseUpgr' + player.id  + pos + zId + \
                   player.upgrade.upgradeType
-        self.udpTransmitter.sendDatagram(message)
+        self.udpClient.sendDatagram(message)
 
     def upgradeDelete(self, player):
         '''Informs the server that the upgrade has been prematurely deleted'''
         pos = struct.pack('ff', *player.pos)
         message = 'DelUpgr' + player.id  + player.upgrade.upgradeType
-        self.udpTransmitter.sendDatagram(message)
+        self.udpClient.sendDatagram(message)
 
     def receiveUpgradeUse(self, player):
         '''Tells the player in question that its upgrade has been deleted'''
         # TODO: um... what?
 
     def requestTransactionInfo(self, team):
-        self.udpTransmitter.sendDatagram('reqTrans' + team.id)
+        self.udpClient.sendDatagram('reqTrans' + team.id)
     
     def validateTransaction(self, team, contributors, totalStars,
                             player, timeLeft, upgrade):
@@ -1294,25 +1335,25 @@ class NetworkClient(object):
                 
     def requestPlayers(self):
         '''Request the list of players currently in the game.'''
-        self.udpTransmitter.sendDatagram('reqPlayers')
+        self.udpClient.sendDatagram('reqPlayers')
         
     def requestZones(self):
         '''Request the current ownership of all zones'''
-        self.udpTransmitter.sendDatagram('reqZones')
+        self.udpClient.sendDatagram('reqZones')
 
     def requestStars(self):
         '''Request the current number of stars that each player has.'''
-        self.udpTransmitter.sendDatagram('reqStars')
+        self.udpClient.sendDatagram('reqStars')
 
     def requestUpgrades(self):
         '''Request the current upgrades that all players have.'''
-        self.udpTransmitter.sendDatagram('reqUpgrades')
+        self.udpClient.sendDatagram('reqUpgrades')
 
     def sendZoneChange(self, player, zone):
         '''Inform the server that a local player has changed zones.'''
         zoneId = struct.pack('!I', zone.id)
         message = 'ZoneCh' + player.id + zoneId
-        self.udpTransmitter.sendDatagram(message)            
+        self.udpClient.sendDatagram(message)            
         
     def sendChat(self, text, sendingPlayer, chatMode = "team"):
         '''Send a chat message to the server. The server will send it out
@@ -1379,7 +1420,7 @@ class NetworkClient(object):
         else:
             message += 'Open'
         message += text
-        self.udpTransmitter.sendDatagram(message)
+        self.udpClient.sendDatagram(message)
         if self.interface:
             self.interface.sentChat(text, sendingPlayer.team, privacy)
 
